/***************************************************************************
 * Copyright 2013 Kieker Project (http://kieker-monitoring.net)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/

package kieker.analysis.plugin.connector.mq;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;

import com.rabbitmq.client.Channel;
import com.rabbitmq.client.Connection;
import com.rabbitmq.client.ConnectionFactory;

import kieker.analysis.IProjectContext;
import kieker.analysis.plugin.annotation.InputPort;
import kieker.analysis.plugin.annotation.OutputPort;
import kieker.analysis.plugin.annotation.Plugin;
import kieker.analysis.plugin.annotation.Property;
import kieker.analysis.plugin.filter.AbstractFilterPlugin;
import kieker.common.configuration.Configuration;
import kieker.common.logging.Log;
import kieker.common.logging.LogFactory;

/**
 * A plugin used for kieker in the cloud.
 * All incoming events are put into a RabbitMQ, but are also passed to an output port that can be used for
 * testing purposes.
 * 
 * @author Santje Finke
 * 
 * @since 1.8
 * 
 */
@Plugin(
		description = "A filter that writes all incoming events into a specified queue on a specified RabbitMQServer",
		outputPorts = {
			@OutputPort(
					name = RabbitMQConnector.OUTPUT_PORT_NAME, eventTypes = { Object.class },
					description = "Provides each incoming object") },
		configuration = {
			@Property(name = RabbitMQConnector.CONFIG_PROPERTY_NAME_PROVIDER, defaultValue = "localhost"),
			@Property(name = RabbitMQConnector.CONFIG_PROPERTY_NAME_QUEUE, defaultValue = "master"),
			@Property(name = RabbitMQConnector.CONFIG_PROPERTY_PASSWORD, defaultValue = "guest"),
			@Property(name = RabbitMQConnector.CONFIG_PROPERTY_USER, defaultValue = "guest")
		})
public class RabbitMQConnector extends AbstractFilterPlugin {

	/**
	 * The name of the input port receiving the incoming events.
	 */
	public static final String INPUT_PORT_NAME_EVENTS = "inputEvents";
	/**
	 * The name of the output port passing the incoming events.
	 */
	public static final String OUTPUT_PORT_NAME = "relayedEvents";
	/**
	 * The name of the property determining the address of the used Server.
	 */
	public static final String CONFIG_PROPERTY_NAME_PROVIDER = "providerUrl";
	/**
	 * The name of the property determining the name of the Queue.
	 */
	public static final String CONFIG_PROPERTY_NAME_QUEUE = "queueName";

	/**
	 * The username that is used to connect to a queue.
	 */
	public static final String CONFIG_PROPERTY_USER = "guest";
	/**
	 * The password that is used to connect to a queue.
	 */
	public static final String CONFIG_PROPERTY_PASSWORD = "guest";

	private static final Log LOG = LogFactory.getLog(RabbitMQConnector.class);

	private final String provider;
	private final String queue;
	private final String password;
	private final String username;
	private Connection connection;
	private Channel channel;
	private final ConnectionFactory factory;

	/**
	 * Creates a new instance of this class using the given parameters.
	 * 
	 * @param configuration
	 *            The configuration for this plugin
	 * 
	 * @deprecated To be removed in Kieker 1.8.
	 */
	@Deprecated
	public RabbitMQConnector(final Configuration configuration) {
		this(configuration, null);
	}

	/**
	 * Creates a new instance of this class using the given parameters.
	 * 
	 * @param configuration
	 *            The configuration for this component.
	 * @param projectContext
	 *            The project context for this component.
	 */
	public RabbitMQConnector(final Configuration configuration, final IProjectContext projectContext) {
		super(configuration, projectContext);
		this.provider = configuration.getStringProperty(CONFIG_PROPERTY_NAME_PROVIDER);
		this.queue = configuration.getStringProperty(CONFIG_PROPERTY_NAME_QUEUE);
		this.username = configuration.getStringProperty(CONFIG_PROPERTY_USER);
		this.password = configuration.getStringProperty(CONFIG_PROPERTY_PASSWORD);
		this.factory = new ConnectionFactory();
		this.factory.setHost(this.provider);
		this.factory.setConnectionTimeout(0);
		this.factory.setUsername(this.username);
		this.factory.setPassword(this.password);
		try {
			this.connection = this.factory.newConnection();
			this.channel = this.connection.createChannel();
			this.channel.queueDeclare(this.queue, false, false, false, null);
		} catch (final IOException e) {
			LOG.info("Error establishing connection", e);
		}
		LOG.info("Sending to destination:" + this.queue + " at " + this.provider + " !\n***\n\n");

	}

	@Override
	public Configuration getCurrentConfiguration() {
		final Configuration configuration = new Configuration();
		configuration.setProperty(CONFIG_PROPERTY_NAME_PROVIDER, this.provider);
		configuration.setProperty(CONFIG_PROPERTY_NAME_QUEUE, this.queue);
		configuration.setProperty(CONFIG_PROPERTY_PASSWORD, this.password);
		configuration.setProperty(CONFIG_PROPERTY_USER, this.username);
		return configuration;
	}

	/**
	 * This method represents the input port of this filter.
	 * 
	 * @param event
	 *            The next event.
	 */
	@InputPort(
			name = INPUT_PORT_NAME_EVENTS,
			eventTypes = { Object.class },
			description = "Receives incoming objects to be forwarded to a queue")
	public final void inputEvent(final Object event) {
		super.deliver(OUTPUT_PORT_NAME, event);
		final ByteArrayOutputStream boas = new ByteArrayOutputStream();
		ObjectOutputStream out;
		try {
			out = new ObjectOutputStream(boas);
			out.writeObject(event);
			out.close();
			final byte[] message = boas.toByteArray();

			if (!this.connection.isOpen() || !this.channel.isOpen()) {
				this.reconnect();
			}

			this.channel.basicPublish("", this.queue, null, message);
		} catch (final IOException e) {
			RabbitMQConnector.LOG.error("Error sending event", e);
		}

	}

	/**
	 * Establishes a connection to a rabbitMQ channel with the current connection informationen.
	 */
	private void reconnect() {
		try {
			this.connection = this.factory.newConnection();
			this.channel = this.connection.createChannel();
			this.channel.queueDeclare(this.queue, false, false, false, null);
		} catch (final IOException e) {
			RabbitMQConnector.LOG.error("Error reestablishing connection", e);
		}
	}

}
