/***************************************************************************
 * Copyright 2013 Kieker Project (http://kieker-monitoring.net)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/

package kieker.analysis.plugin.filter.flow;

import java.io.Serializable;
import java.util.Comparator;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.TimeUnit;

import kieker.analysis.IProjectContext;
import kieker.analysis.plugin.annotation.InputPort;
import kieker.analysis.plugin.annotation.OutputPort;
import kieker.analysis.plugin.annotation.Plugin;
import kieker.analysis.plugin.annotation.Property;
import kieker.analysis.plugin.filter.AbstractFilterPlugin;
import kieker.common.configuration.Configuration;
import kieker.common.logging.Log;
import kieker.common.logging.LogFactory;
import kieker.common.record.flow.IFlowRecord;
import kieker.common.record.flow.trace.AbstractTraceEvent;
import kieker.common.record.flow.trace.WorkflowRecord;
import kieker.common.record.flow.trace.WorkflowTrace;

/**
 * This filter merges partial traces with the same trace id into one complete trace.
 * Incomplete traces will be delivered after a specified timeout.
 * 
 * @author Florian Biss, Soeren Mahmens, Bjoern Weissenfels
 * 
 * @since 1.8
 */
@Plugin(name = "Final Trace Reconstruction Filter (Workflow)",
		description = "This filter merges partial WorkflowTraces into complete traces.",
		outputPorts = {
			@OutputPort(name = WorkflowRecordFinalTraceReconstructionFilter.OUTPUT_PORT_NAME_VALID_TRACES,
					description = "Forwards valid traces",
					eventTypes = { WorkflowTrace.class }),
			@OutputPort(name = WorkflowRecordFinalTraceReconstructionFilter.OUTPUT_PORT_NAME_INVALID_TRACES,
					description = "Forwards invalid traces",
					eventTypes = { WorkflowTrace.class })
		},
		configuration = {
			@Property(name = WorkflowRecordFinalTraceReconstructionFilter.CONFIG_PROPERTY_NAME_TIMEUNIT,
					defaultValue = WorkflowRecordFinalTraceReconstructionFilter.CONFIG_PROPERTY_VALUE_TIMEUNIT),
			@Property(name = WorkflowRecordFinalTraceReconstructionFilter.CONFIG_PROPERTY_NAME_MAX_TRACE_TIMEOUT,
					defaultValue = WorkflowRecordFinalTraceReconstructionFilter.CONFIG_PROPERTY_VALUE_MAX_TIME),
			@Property(name = WorkflowRecordFinalTraceReconstructionFilter.CONFIG_PROPERTY_NAME_MAX_TRACE_DURATION,
					defaultValue = WorkflowRecordFinalTraceReconstructionFilter.CONFIG_PROPERTY_VALUE_MAX_TIME)
		})
public class WorkflowRecordFinalTraceReconstructionFilter extends AbstractFilterPlugin {
	/**
	 * The name of the output port delivering the valid traces.
	 */
	public static final String OUTPUT_PORT_NAME_VALID_TRACES = "validTraces";
	/**
	 * The name of the output port delivering the valid traces.
	 */
	public static final String OUTPUT_PORT_NAME_INVALID_TRACES = "invalidTraces";
	/**
	 * The name of the input port receiving the trace records.
	 */
	public static final String INPUT_PORT_NAME_PARTIAL_TRACES = "partialTraces";

	/**
	 * The name of the time trigger input port.
	 */
	public static final String INPUT_PORT_NAME_TIME_EVENT = "timestamp";

	/**
	 * The name of the property determining the time unit.
	 */
	public static final String CONFIG_PROPERTY_NAME_TIMEUNIT = "timeunit";
	/**
	 * The name of the property determining the maximal trace duration, the time
	 * this filter waits for new partial traces of a trace.
	 */
	public static final String CONFIG_PROPERTY_NAME_MAX_TRACE_DURATION = "maxTraceDuration";
	/**
	 * The name of the property determining the maximal trace timeout.
	 */
	public static final String CONFIG_PROPERTY_NAME_MAX_TRACE_TIMEOUT = "maxTraceTimeout";
	/**
	 * The default value of the properties for the maximal trace duration and timeout.
	 */
	public static final String CONFIG_PROPERTY_VALUE_MAX_TIME = "9223372036854775807"; // String.valueOf(Long.MAX_VALUE)
	/**
	 * The default value of the time unit property (nanoseconds).
	 */
	public static final String CONFIG_PROPERTY_VALUE_TIMEUNIT = "NANOSECONDS"; // TimeUnit.NANOSECONDS.name()

	private static final Log LOG = LogFactory.getLog(WorkflowRecordFinalTraceReconstructionFilter.class);

	private final TimeUnit timeunit;
	private final long maxTraceDuration;
	private final long maxTraceTimeout;

	private final ConcurrentMap<Long, TraceBuffer> traceId2trace;

	/**
	 * Creates a new instance of this class using the given parameters.
	 * 
	 * @param configuration
	 *            The configuration for this component.
	 * @param projectContext
	 *            The project context for this component.
	 */
	public WorkflowRecordFinalTraceReconstructionFilter(final Configuration configuration, final IProjectContext projectContext) {
		super(configuration, projectContext);

		if (null != projectContext) { // TODO #819 remove non-null check and else case in Kieker 1.8 //NOCS
			final String recordTimeunitProperty = projectContext.getProperty(IProjectContext.CONFIG_PROPERTY_NAME_RECORDS_TIME_UNIT);
			TimeUnit recordTimeunit;
			try {
				recordTimeunit = TimeUnit.valueOf(recordTimeunitProperty);
			} catch (final IllegalArgumentException ex) { // already caught in AnalysisController, should never happen
				LOG.warn(recordTimeunitProperty + " is no valid TimeUnit! Using NANOSECONDS instead.");
				recordTimeunit = TimeUnit.NANOSECONDS;
			}
			this.timeunit = recordTimeunit;
		} else {
			this.timeunit = TimeUnit.NANOSECONDS;
		}

		final String configTimeunitProperty = configuration.getStringProperty(CONFIG_PROPERTY_NAME_TIMEUNIT);
		TimeUnit configTimeunit;
		try {
			configTimeunit = TimeUnit.valueOf(configTimeunitProperty);
		} catch (final IllegalArgumentException ex) {
			LOG.warn(configTimeunitProperty + " is no valid TimeUnit! Using inherited value of " + this.timeunit.name() + " instead.");
			configTimeunit = this.timeunit;
		}

		this.maxTraceDuration = this.timeunit.convert(configuration.getLongProperty(CONFIG_PROPERTY_NAME_MAX_TRACE_DURATION), configTimeunit);
		this.maxTraceTimeout = this.timeunit.convert(configuration.getLongProperty(CONFIG_PROPERTY_NAME_MAX_TRACE_TIMEOUT), configTimeunit);
		this.traceId2trace = new ConcurrentHashMap<Long, TraceBuffer>();
	}

	/**
	 * This method is the input port for the timeout.
	 * 
	 * @param timestamp
	 *            The actual nanotime
	 */
	@InputPort(
			name = INPUT_PORT_NAME_TIME_EVENT,
			description = "Input port for periodic a time signal",
			eventTypes = { Long.class })
	public void newEvent(final Long timestamp) {
		synchronized (this) {
			for (final TraceBuffer traceBuffer : this.traceId2trace.values()) {
				final long timeSinceLastRecord = timestamp - traceBuffer.getMaxLoggingTimestamp();
				final long timeSinceFirstRecord = timestamp - traceBuffer.getMinLoggingTimestamp();
				if ((timeSinceLastRecord >= this.maxTraceTimeout) || (timeSinceFirstRecord >= this.maxTraceDuration)) { // max duration or timeout is gone
					super.deliver(OUTPUT_PORT_NAME_INVALID_TRACES, traceBuffer.toWorkflowTrace());
					// Concurrent modification allowed by ConcurrentMap
					this.traceId2trace.remove(traceBuffer.getTraceId());
				}
			}
		}
	}

	/**
	 * This method is the input port for the new events for this filter.
	 * 
	 * @param record
	 *            The new record to handle.
	 */
	@InputPort(
			name = INPUT_PORT_NAME_PARTIAL_TRACES,
			description = "Input port for partial traces from WorkflowRecordPartialTraceReconstructionFilters",
			eventTypes = { WorkflowTrace.class })
	public void newEvent(final IFlowRecord record) {
		final Long traceId;
		TraceBuffer traceBuffer;
		final WorkflowTrace trace;

		if (record instanceof WorkflowTrace) {
			trace = (WorkflowTrace) record;

			if (trace.isComplete()) {
				super.deliver(OUTPUT_PORT_NAME_VALID_TRACES, trace); // Nothing to do here
				return;
			} else {
				traceId = trace.getTraceId();
				synchronized (this) {
					traceBuffer = this.getTraceBuffer(traceId);
					traceBuffer.insertTrace(trace);
				}
			}
		} else {
			LOG.error("Invalid input type at " + INPUT_PORT_NAME_PARTIAL_TRACES
					+ " in WorkflowRecordFinalTraceReconstructionFilter");
			return; // invalid type
		}

		synchronized (this) {
			if (traceBuffer.isComplete()) {
				this.traceId2trace.remove(traceId);
				super.deliver(OUTPUT_PORT_NAME_VALID_TRACES, traceBuffer.toWorkflowTrace());
			}
		}
	}

	private TraceBuffer getTraceBuffer(final Long traceId) {
		TraceBuffer traceBuffer;
		traceBuffer = this.traceId2trace.get(traceId);
		if (traceBuffer == null) { // first record for this id!
			synchronized (this) {
				traceBuffer = this.traceId2trace.get(traceId);

				if (traceBuffer == null) { // NOCS (DCL)
					final TraceBuffer newTraceBuffer = new TraceBuffer(this.timeunit);
					traceBuffer = this.traceId2trace.put(traceId, newTraceBuffer);
					if (traceBuffer == null) {
						traceBuffer = newTraceBuffer;
					}
				}
			}
		}
		return traceBuffer;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void terminate(final boolean error) {
		synchronized (this) {
			for (final TraceBuffer traceBuffer : this.traceId2trace.values()) {
				super.deliver(OUTPUT_PORT_NAME_INVALID_TRACES, traceBuffer.toWorkflowTrace());
			}
			this.traceId2trace.clear();
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public Configuration getCurrentConfiguration() {
		final Configuration configuration = new Configuration();
		configuration.setProperty(CONFIG_PROPERTY_NAME_TIMEUNIT, this.timeunit.name());
		configuration.setProperty(CONFIG_PROPERTY_NAME_MAX_TRACE_DURATION, String.valueOf(this.maxTraceDuration));
		configuration.setProperty(CONFIG_PROPERTY_NAME_MAX_TRACE_TIMEOUT, String.valueOf(this.maxTraceTimeout));
		return configuration;
	}

	/**
	 * Buffer for events from partial traces that will be turned into a single trace.
	 * 
	 * @author Florian Biss
	 */
	private static final class TraceBuffer {
		private static final Log LOG = LogFactory.getLog(TraceBuffer.class);
		private static final Comparator<AbstractTraceEvent> COMPARATOR = new TraceEventComperator();

		private final SortedSet<AbstractTraceEvent> events = new TreeSet<AbstractTraceEvent>(COMPARATOR);

		private boolean damaged;
		private int maxOrderIndex = -1;

		private final TimeUnit timeunit;

		private long minLoggingTimestamp = Long.MAX_VALUE;
		private long maxLoggingTimestamp = -1;

		private long traceId = -1;
		private boolean hasStart;
		private boolean hasEnd;

		/**
		 * Creates a new buffer.
		 * 
		 * @param timeunit
		 *            TimetUnit used for logging timestamps.
		 */
		public TraceBuffer(final TimeUnit timeunit) {
			this.timeunit = timeunit;
		}

		/**
		 * Store all events of a partial trace in the buffer.
		 * 
		 * @param paritalTrace
		 *            A partial trace
		 */
		public void insertTrace(final WorkflowTrace paritalTrace) {
			final long myTraceId = paritalTrace.getTraceId();
			// Time information in partial traces are old (partial traces timed out on a worker),
			// use current system time instead.
			final long loggingTimestamp = this.timeunit.convert(System.nanoTime(), TimeUnit.NANOSECONDS);

			synchronized (this) {

				if (this.traceId == -1) {
					this.traceId = myTraceId;
					this.minLoggingTimestamp = loggingTimestamp;
				} else if (this.traceId != myTraceId) {
					LOG.error("Invalid traceId! Expected: " + this.traceId + " but found: " + myTraceId + " in trace " + paritalTrace.toString());
					this.damaged = true;
				}

				this.maxLoggingTimestamp = loggingTimestamp;

				final WorkflowRecord[] newEvents = paritalTrace.getTraceEvents();

				for (final WorkflowRecord event : newEvents) {
					this.insertEvent(event);
				}

				if (paritalTrace.isDamaged()) {
					this.damaged = true;
				}
			}
		}

		private void insertEvent(final WorkflowRecord event) {
			final int orderIndex = event.getOrderIndex();
			if (orderIndex > this.maxOrderIndex) {
				this.maxOrderIndex = orderIndex;
			}

			if (event.isStart()) {
				if (this.hasStart) {
					LOG.error("Duplicate start event! TraceId: " + this.traceId + " Event: " + event.toString());
					this.damaged = true;
				}
				this.hasStart = true;
			}

			if (event.isEnd()) {
				if (this.hasEnd) {
					LOG.error("Duplicate end event! TraceId: " + this.traceId + " Event: " + event.toString());
					this.damaged = true;
				}
				this.hasEnd = true;
			}

			if (!this.events.add(event)) {
				LOG.error("Duplicate entry for orderIndex " + orderIndex + " with traceId " + this.traceId);
				this.damaged = true;
			}
		}

		/**
		 * @return
		 *         <code>true</code> if all records are present and the trace has a start and an end record.
		 */
		public boolean isComplete() {
			synchronized (this) {
				return ((this.maxOrderIndex + 1) == this.events.size()) && !this.events.isEmpty()
						&& this.hasEnd && this.hasStart && !this.damaged;
			}
		}

		/**
		 * @return <code>true</code> if the trace in this buffer is damaged.
		 */
		public boolean isDamaged() {
			synchronized (this) {
				return this.damaged;
			}
		}

		/**
		 * @return The trace id
		 */
		public long getTraceId() {
			synchronized (this) {
				return this.traceId;
			}
		}

		/**
		 * Process this buffer into a WorkflowTrace containing all buffered events.
		 * 
		 * @return A new WorkflowTrace
		 */
		public WorkflowTrace toWorkflowTrace() {
			synchronized (this) {
				return new WorkflowTrace(this.events.toArray(new WorkflowRecord[this.events.size()]),
						this.isComplete(), this.isDamaged());
			}
		}

		/**
		 * {@inheritDoc}
		 */
		@Override
		public String toString() {
			return this.toWorkflowTrace().toString();
		}

		/**
		 * @return Insertion timestamp of the latest event in this buffer
		 */
		public long getMaxLoggingTimestamp() {
			synchronized (this) {
				return this.maxLoggingTimestamp;
			}
		}

		/**
		 * @return Insertion timestamp of the first event in this buffer
		 */
		public long getMinLoggingTimestamp() {
			synchronized (this) {
				return this.minLoggingTimestamp;
			}
		}

		/**
		 * Compares two trace events by their order index.
		 * 
		 * @author Jan Waller
		 */
		private static final class TraceEventComperator implements Comparator<AbstractTraceEvent>, Serializable {
			private static final long serialVersionUID = 89207356648232517L;

			/**
			 * Creates a new instance of this class.
			 */
			public TraceEventComperator() {
				// default empty constructor
			}

			/**
			 * {@inheritDoc}
			 */
			public int compare(final AbstractTraceEvent o1, final AbstractTraceEvent o2) {
				return o1.getOrderIndex() - o2.getOrderIndex();
			}
		}
	}
}
