/***************************************************************************
 * Copyright 2013 Kieker Project (http://kieker-monitoring.net)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/

package kieker.analysis.plugin.filter.flow;

import java.io.Serializable;
import java.util.Comparator;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;

import kieker.analysis.IProjectContext;
import kieker.analysis.plugin.annotation.InputPort;
import kieker.analysis.plugin.annotation.OutputPort;
import kieker.analysis.plugin.annotation.Plugin;
import kieker.analysis.plugin.annotation.Property;
import kieker.analysis.plugin.filter.AbstractFilterPlugin;
import kieker.common.configuration.Configuration;
import kieker.common.logging.Log;
import kieker.common.logging.LogFactory;
import kieker.common.record.flow.IFlowRecord;
import kieker.common.record.flow.trace.AbstractTraceEvent;
import kieker.common.record.flow.trace.WorkflowRecord;
import kieker.common.record.flow.trace.WorkflowTrace;

/**
 * This filter collects WorkflowRecords and constructs ordered traces of them.
 * 
 * 
 * @author Florian Biss, Soeren Mahmens, Bjoern Weissenfels
 * 
 * @since 1.8
 */
@Plugin(name = "Partial Trace Reconstruction Filter (Workflow)",
		description = "This filter bundles WorkflowRecords into a trace",
		outputPorts = {
			@OutputPort(name = WorkflowRecordPartialTraceReconstructionFilter.OUTPUT_PORT_NAME_PARTIAL_TRACES,
					description = "Forwards the constructed partial and complete traces",
					eventTypes = { WorkflowTrace.class })
		},
		configuration = {
			@Property(name = WorkflowRecordPartialTraceReconstructionFilter.CONFIG_PROPERTY_NAME_MAX_TRACE_TIMEOUT_NS,
					defaultValue = WorkflowRecordPartialTraceReconstructionFilter.CONFIG_PROPERTY_VALUE_MAX_TRACE_TIMEOUT_NS),
			@Property(name = WorkflowRecordPartialTraceReconstructionFilter.CONFIG_PROPERTY_NAME_MAX_TRACE_DURATION_NS,
					defaultValue = WorkflowRecordPartialTraceReconstructionFilter.CONFIG_PROPERTY_VALUE_MAX_TRACE_DURATION_NS)
		})
public class WorkflowRecordPartialTraceReconstructionFilter extends AbstractFilterPlugin {
	/**
	 * The name of the output port delivering the valid traces.
	 */
	public static final String OUTPUT_PORT_NAME_PARTIAL_TRACES = "partialTraces";
	/**
	 * The name of the input port receiving the trace records.
	 */
	public static final String INPUT_PORT_NAME_WORKFLOW_RECORDS = "workflowRecords";

	/**
	 * The name of the input port receiving the time stamps.
	 */
	public static final String INPUT_PORT_NAME_TIME_EVENT = "timestamp";
	/**
	 * The name of the property determining the maximal trace timeout.
	 */
	public static final String CONFIG_PROPERTY_NAME_MAX_TRACE_TIMEOUT_NS = "maxTraceTimeout";

	/**
	 * The default value of the property determining the maximal trace timeout.
	 */
	public static final String CONFIG_PROPERTY_VALUE_MAX_TRACE_TIMEOUT_NS = "1000000000"; // 1 seconds

	/**
	 * The name of the property determining the maximal trace duration.
	 */
	public static final String CONFIG_PROPERTY_NAME_MAX_TRACE_DURATION_NS = "maxTraceDuration";

	/**
	 * The default value of the property determining the maximal trace duration.
	 */
	public static final String CONFIG_PROPERTY_VALUE_MAX_TRACE_DURATION_NS = "5000000000"; // 5 seconds

	private static final Log LOG = LogFactory.getLog(WorkflowRecordPartialTraceReconstructionFilter.class);

	private final ConcurrentMap<Long, TraceBuffer> traceId2trace;
	private final long maxTraceTimeout;
	private final long maxTraceDuration;

	/**
	 * Creates a new instance of this class using the given parameters.
	 * 
	 * @param configuration
	 *            The configuration for this component.
	 * @param projectContext
	 *            The project context for this component.
	 */
	public WorkflowRecordPartialTraceReconstructionFilter(final Configuration configuration, final IProjectContext projectContext) {
		super(configuration, projectContext);

		this.traceId2trace = new ConcurrentHashMap<Long, TraceBuffer>();

		this.maxTraceTimeout = configuration.getLongProperty(CONFIG_PROPERTY_NAME_MAX_TRACE_TIMEOUT_NS);
		this.maxTraceDuration = configuration.getLongProperty(CONFIG_PROPERTY_NAME_MAX_TRACE_DURATION_NS);

	}

	/**
	 * This method is the input port for the timeout.
	 * 
	 * @param timestamp
	 *            The actual nanotime
	 */
	@InputPort(
			name = INPUT_PORT_NAME_TIME_EVENT,
			description = "Input port for periodic a time signal",
			eventTypes = { Long.class })
	public void newEvent(final Long timestamp) {
		synchronized (this) {
			for (final TraceBuffer traceBuffer : this.traceId2trace.values()) {
				final long timeSinceLastRecord = timestamp - traceBuffer.getMaxLoggingTimestamp();
				final long timeSinceFirstRecord = timestamp - traceBuffer.getMinLoggingTimestamp();
				if ((timeSinceLastRecord >= this.maxTraceTimeout) || (timeSinceFirstRecord >= this.maxTraceDuration)) { // max timeout or duration is gone
					this.traceId2trace.remove(traceBuffer.getTraceID());
					super.deliver(OUTPUT_PORT_NAME_PARTIAL_TRACES, traceBuffer.toWorkflowTrace());
				}
			}
		}
	}

	/**
	 * This method is the input port for the new events for this filter.
	 * 
	 * @param record
	 *            The new record to handle.
	 */
	@InputPort(
			name = INPUT_PORT_NAME_WORKFLOW_RECORDS,
			description = "Input port for WorkflowRecords",
			eventTypes = { WorkflowRecord.class })
	public void newEvent(final IFlowRecord record) {

		final Long traceId;
		TraceBuffer traceBuffer;

		if (record instanceof WorkflowRecord) {
			final WorkflowRecord traceRecord = (WorkflowRecord) record;
			traceId = traceRecord.getTraceId();
			traceBuffer = this.getTraceBuffer(traceId);

			synchronized (this) {
				traceBuffer.insertEvent(traceRecord);

				if (traceBuffer.hasEnd()) {
					this.traceId2trace.remove(traceId);
					super.deliver(OUTPUT_PORT_NAME_PARTIAL_TRACES, traceBuffer.toWorkflowTrace());
				}
			}

		} else {
			LOG.error("Invalid event type in WorkflowRecordPartialTraceReconstructionFilter");
			return; // invalid type which should not happen due to the specified eventTypes
		}

	}

	private TraceBuffer getTraceBuffer(final Long traceId) {
		TraceBuffer traceBuffer;
		traceBuffer = this.traceId2trace.get(traceId);
		if (traceBuffer == null) { // first record for this id!
			synchronized (this) {
				traceBuffer = this.traceId2trace.get(traceId);

				if (traceBuffer == null) { // NOCS (DCL)
					final TraceBuffer newTraceBuffer = new TraceBuffer();
					traceBuffer = this.traceId2trace.put(traceId, newTraceBuffer);
					if (traceBuffer == null) {
						traceBuffer = newTraceBuffer;
					}
				}
			}
		}
		return traceBuffer;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void terminate(final boolean error) {
		this.deliverAllBuffer();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public Configuration getCurrentConfiguration() {
		final Configuration configuration = new Configuration();
		configuration.setProperty(CONFIG_PROPERTY_NAME_MAX_TRACE_TIMEOUT_NS, String.valueOf(this.maxTraceTimeout));
		return configuration;
	}

	/**
	 * This method delivers all traceBuffer and clears the buffer.
	 */
	private void deliverAllBuffer() {
		synchronized (this) {
			for (final TraceBuffer traceBuffer : this.traceId2trace.values()) {
				super.deliver(OUTPUT_PORT_NAME_PARTIAL_TRACES, traceBuffer.toWorkflowTrace());
			}
			this.traceId2trace.clear();
		}
	}

	/**
	 * Buffer for records that will be bundled into a single trace.
	 * 
	 * @author Florian Biss
	 */
	private static final class TraceBuffer {
		private static final Log LOG = LogFactory.getLog(TraceBuffer.class);
		private static final Comparator<AbstractTraceEvent> COMPARATOR = new TraceEventComperator();

		private final SortedSet<WorkflowRecord> events = new TreeSet<WorkflowRecord>(COMPARATOR);

		private boolean damaged;
		private int maxOrderIndex = -1;

		private long minLoggingTimestamp = Long.MAX_VALUE;
		private long maxLoggingTimestamp = -1;

		private long traceId = -1;
		private boolean hasStart;
		private boolean ended;

		/**
		 * Creates a new buffer.
		 */
		public TraceBuffer() {
			// default empty constructor
		}

		/**
		 * @return The trace id
		 */
		public Object getTraceID() {
			synchronized (this) {
				return this.traceId;
			}
		}

		/**
		 * Insert a new event into buffer.
		 * 
		 * @param event
		 *            New event
		 */
		public void insertEvent(final WorkflowRecord event) {
			final long myTraceId = event.getTraceId();
			synchronized (this) {

				final long currentTime = System.nanoTime();
				if (this.traceId == -1) {
					this.traceId = myTraceId;
					this.minLoggingTimestamp = currentTime;
				} else if (this.traceId != myTraceId) {
					LOG.error("Invalid traceId! Expected: " + this.traceId + " but found: " + myTraceId + " in event " + event.toString());
					this.damaged = true;
				}

				this.maxLoggingTimestamp = currentTime;

				final int orderIndex = event.getOrderIndex();
				if (orderIndex > this.maxOrderIndex) {
					this.maxOrderIndex = orderIndex;
				}

				if (event.isStart()) {
					if (this.hasStart) {
						LOG.error("Duplicate start event! TraceId: " + this.traceId + " Event: " + event.toString());
						this.damaged = true;
					}
					this.hasStart = true;
				}

				if (event.isEnd()) {
					if (this.ended) {
						LOG.error("Duplicate end event! TraceId: " + this.traceId + " Event: " + event.toString());
						this.damaged = true;
					}
					this.ended = true;
				}

				if (!this.events.add(event)) {
					LOG.error("Duplicate entry for orderIndex " + orderIndex + " with traceId " + myTraceId);
					this.damaged = true;
				}
			}
		}

		/**
		 * @return <code>true</code> if this buffer contains a trace end record
		 */
		public boolean hasEnd() {
			synchronized (this) {
				return this.ended;
			}
		}

		/**
		 * @return <code>true</code> if this buffer contains all records of a trace, including start and end record
		 */
		public boolean isComplete() {
			synchronized (this) {
				return ((this.maxOrderIndex + 1) == this.events.size())
						&& !this.events.isEmpty() && this.ended && this.hasStart;
			}
		}

		/**
		 * @return <code>true</code> if the trace in this buffer is damaged
		 */
		public boolean isDamaged() {
			synchronized (this) {
				return this.damaged;
			}
		}

		/**
		 * Process buffer into a trace.
		 * 
		 * @return
		 *         A new trace containing the buffered events
		 */
		public WorkflowTrace toWorkflowTrace() {
			return new WorkflowTrace(this.events.toArray(new WorkflowRecord[this.events.size()]),
					this.isComplete(), this.isDamaged());
		}

		/**
		 * @return Youngest time stamp in trace
		 */
		public long getMaxLoggingTimestamp() {
			synchronized (this) {
				return this.maxLoggingTimestamp;
			}
		}

		/**
		 * @return Oldest time stamp in trace
		 */
		public long getMinLoggingTimestamp() {
			synchronized (this) {
				return this.minLoggingTimestamp;
			}
		}

		/**
		 * @author Jan Waller
		 */
		private static final class TraceEventComperator implements Comparator<AbstractTraceEvent>, Serializable {
			private static final long serialVersionUID = 89207356648232517L;

			/**
			 * Creates a new instance of this class.
			 */
			public TraceEventComperator() {
				// default empty constructor
			}

			public int compare(final AbstractTraceEvent o1, final AbstractTraceEvent o2) {
				return o1.getOrderIndex() - o2.getOrderIndex();
			}
		}
	}
}
