/***************************************************************************
 * Copyright 2013 Kieker Project (http://kieker-monitoring.net)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/

package kieker.analysis.plugin.filter.flow;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.concurrent.TimeUnit;

import kieker.analysis.IProjectContext;
import kieker.analysis.plugin.annotation.InputPort;
import kieker.analysis.plugin.annotation.OutputPort;
import kieker.analysis.plugin.annotation.Plugin;
import kieker.analysis.plugin.annotation.Property;
import kieker.analysis.plugin.filter.AbstractFilterPlugin;
import kieker.common.configuration.Configuration;
import kieker.common.logging.Log;
import kieker.common.logging.LogFactory;
import kieker.common.record.flow.IFlowRecord;
import kieker.common.record.flow.trace.WorkflowRecord;
import kieker.common.record.flow.trace.WorkflowTrace;
import kieker.common.util.StatisticInformation;

/**
 * This filter collects incoming traces for a specified amount of time.
 * Any traces representing the same series of events will be used to calculate
 * statistic informations like the average runtime of this kind of trace.
 * Only one specimen of these traces containing this information will be forwarded
 * from this filter.
 * 
 * Statistic outliers regarding the runtime of the trace will be treated special
 * and therefore send out as they are and will not be mixed with others.
 * 
 * 
 * @author Florian Biss
 * 
 * @since 1.8
 */

@Plugin(description = "This filter tries to agglomerate similar WorkflowTraces into a single trace.",
		outputPorts = {
			@OutputPort(name = WorkflowRecordTraceAgglomerationFilter.OUTPUT_PORT_NAME_TRACES,
					description = "Output port for the processed traces",
					eventTypes = { WorkflowTrace.class })
		},
		configuration = {
			@Property(name = WorkflowRecordTraceAgglomerationFilter.CONFIG_PROPERTY_NAME_TIMEUNIT,
					defaultValue = WorkflowRecordTraceAgglomerationFilter.CONFIG_PROPERTY_VALUE_TIMEUNIT),
			@Property(name = WorkflowRecordTraceAgglomerationFilter.CONFIG_PROPERTY_NAME_MAX_COLLECTION_DURATION,
					defaultValue = WorkflowRecordTraceAgglomerationFilter.CONFIG_PROPERTY_VALUE_MAX_COLLECTION_DURATION),
			@Property(name = WorkflowRecordTraceAgglomerationFilter.CONFIG_PROPERTY_NAME_MAX_DEVIATION,
					defaultValue = WorkflowRecordTraceAgglomerationFilter.CONFIG_PROPERTY_VALUE_MAX_DEVIATION)
		})
public class WorkflowRecordTraceAgglomerationFilter extends AbstractFilterPlugin {
	/**
	 * The name of the output port delivering the valid traces.
	 */
	public static final String OUTPUT_PORT_NAME_TRACES = "tracesOut";

	/**
	 * The name of the input port receiving the trace records.
	 */
	public static final String INPUT_PORT_NAME_TRACES = "tracesIn";

	/**
	 * The name of the property determining the time unit.
	 */
	public static final String CONFIG_PROPERTY_NAME_TIMEUNIT = "timeunit";

	/**
	 * Clock input for timeout handling.
	 */
	public static final String INPUT_PORT_NAME_TIME_EVENT = "timestamp";

	/**
	 * The default value of the time unit property (nanoseconds).
	 */
	public static final String CONFIG_PROPERTY_VALUE_TIMEUNIT = "NANOSECONDS"; // TimeUnit.NANOSECONDS.name()

	/**
	 * The name of the property determining the maximal trace timeout.
	 */
	public static final String CONFIG_PROPERTY_NAME_MAX_COLLECTION_DURATION = "maxCollectionDuration";

	/**
	 * The default value of the property determining the maximal trace timeout.
	 */
	public static final String CONFIG_PROPERTY_VALUE_MAX_COLLECTION_DURATION = "5000000000";

	/**
	 * The name of the property determining the maximal runtime deviation factor.
	 * 
	 * Outliers are indicated by <code>|runtime - averageRuntime| > deviationFactor * standardDeviation</code>.
	 * Use negative number to agglomerate all traces.
	 */
	public static final String CONFIG_PROPERTY_NAME_MAX_DEVIATION = "maxDeviation";

	/**
	 * The default value of the property determining the maximal runtime deviation factor.
	 * Default is two standard deviations.
	 */
	public static final String CONFIG_PROPERTY_VALUE_MAX_DEVIATION = "2";

	private static final Log LOG = LogFactory.getLog(WorkflowRecordTraceAgglomerationFilter.class);

	private final TimeUnit timeunit;
	private final long maxCollectionDuration;
	private final long maxDeviation;

	private final Map<WorkflowTrace, TraceAgglomerationBuffer> trace2buffer;

	/**
	 * Creates a new instance of this class using the given parameters.
	 * 
	 * @param configuration
	 *            The configuration for this component.
	 * @param projectContext
	 *            The project context for this component.
	 */
	public WorkflowRecordTraceAgglomerationFilter(final Configuration configuration, final IProjectContext projectContext) {
		super(configuration, projectContext);

		if (null != projectContext) { // TODO #819 remove non-null check and else case in Kieker 1.8 //NOCS
			final String recordTimeunitProperty = projectContext.getProperty(IProjectContext.CONFIG_PROPERTY_NAME_RECORDS_TIME_UNIT);
			TimeUnit recordTimeunit;
			try {
				recordTimeunit = TimeUnit.valueOf(recordTimeunitProperty);
			} catch (final IllegalArgumentException ex) { // already caught in AnalysisController, should never happen
				LOG.warn(recordTimeunitProperty + " is no valid TimeUnit! Using NANOSECONDS instead.");
				recordTimeunit = TimeUnit.NANOSECONDS;
			}
			this.timeunit = recordTimeunit;
		} else {
			this.timeunit = TimeUnit.NANOSECONDS;
		}
		this.maxDeviation = configuration.getLongProperty(CONFIG_PROPERTY_NAME_MAX_DEVIATION);
		final String configTimeunitProperty = configuration.getStringProperty(CONFIG_PROPERTY_NAME_TIMEUNIT);
		TimeUnit configTimeunit;
		try {
			configTimeunit = TimeUnit.valueOf(configTimeunitProperty);
		} catch (final IllegalArgumentException ex) {
			LOG.warn(configTimeunitProperty + " is no valid TimeUnit! Using inherited value of " + this.timeunit.name() + " instead.");
			configTimeunit = this.timeunit;
		}
		this.maxCollectionDuration = this.timeunit.convert(configuration.getLongProperty(CONFIG_PROPERTY_NAME_MAX_COLLECTION_DURATION), configTimeunit);
		this.trace2buffer = new TreeMap<WorkflowTrace, TraceAgglomerationBuffer>(new TraceComperator());
	}

	/**
	 * This method is the input port for incoming traces.
	 * 
	 * @param record
	 *            A WorkflowTrace
	 */
	@InputPort(
			name = INPUT_PORT_NAME_TRACES,
			description = "Collect identical traces and agglomerate them.",
			eventTypes = { WorkflowTrace.class })
	public void newEvent(final IFlowRecord event) {
		final WorkflowTrace trace;
		synchronized (this) {
			if (event instanceof WorkflowTrace) {
				trace = (WorkflowTrace) event;

				if (!trace.isComplete() || trace.isDamaged()) {
					super.deliver(OUTPUT_PORT_NAME_TRACES, trace); // Incomplete or damaged? Nothing to do here.
					return;
				} else {
					this.insertIntoBuffer(trace);
				}
			} else {
				LOG.error("Invalid input type at " + OUTPUT_PORT_NAME_TRACES
						+ " in WorkflowRecordTraceAgglomerationFilter");
				return; // invalid type
			}
		}
	}

	/**
	 * Inserts a WorkflowTrace into the buffer.
	 * 
	 * @param trace
	 *            The WorkflowTrace that will be inserted
	 */
	private void insertIntoBuffer(final WorkflowTrace trace) {

		TraceAgglomerationBuffer traceBuffer;
		final long timestamp;
		timestamp = this.timeunit.convert(System.nanoTime(), TimeUnit.NANOSECONDS);
		traceBuffer = this.trace2buffer.get(trace);

		if (traceBuffer == null) { // first record for this id!
			synchronized (this) {
				traceBuffer = this.trace2buffer.get(trace);

				if (traceBuffer == null) { // NOCS (DCL)
					traceBuffer = new TraceAgglomerationBuffer(timestamp, this.maxDeviation);
					this.trace2buffer.put(trace, traceBuffer);
				}

			}
		}

		traceBuffer.insertTrace(trace);
	}

	/**
	 * This method is the input port for the timeout.
	 * 
	 * @param timestamp
	 *            The current nanotime
	 */
	@InputPort(
			name = INPUT_PORT_NAME_TIME_EVENT,
			description = "Time signal for timeouts",
			eventTypes = { Long.class })
	public void newEvent(final Long timestamp) {
		synchronized (this) {
			this.processTimeoutQueue(timestamp);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void terminate(final boolean error) {
		synchronized (this) {
			// Avoid ConcurrentModificationException, deliverBuffer will remove the buffer from trace2buffer.
			final List<TraceAgglomerationBuffer> buffers = new LinkedList<TraceAgglomerationBuffer>(this.trace2buffer.values());
			for (final TraceAgglomerationBuffer traceBuffer : buffers) {
				this.deliverBuffer(traceBuffer);
			}
			this.trace2buffer.clear();
		}
	}

	private void processTimeoutQueue(final long timestamp) {
		final long bufferTimeout = timestamp - this.maxCollectionDuration;
		// Avoid ConcurrentModificationException, deliverBuffer will remove the buffer from trace2buffer.
		final List<TraceAgglomerationBuffer> buffers = new LinkedList<TraceAgglomerationBuffer>(this.trace2buffer.values());
		for (final TraceAgglomerationBuffer traceBuffer : buffers) {
			if (traceBuffer.getBufferCreatedTimestamp() <= bufferTimeout) {
				this.deliverBuffer(traceBuffer);
			}
		}
	}

	/**
	 * Deliver and remove a buffer.
	 * 
	 * @param traceBuffer
	 */
	private void deliverBuffer(final TraceAgglomerationBuffer traceBuffer) {
		final List<WorkflowTrace> traces = traceBuffer.processBuffer();
		for (final WorkflowTrace trace : traces) {
			super.deliver(OUTPUT_PORT_NAME_TRACES, trace);
		}
		// Kill the buffer
		if (this.trace2buffer.remove(traces.get(0)) == null) {
			LOG.warn("Removal of buffer failed.");
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public Configuration getCurrentConfiguration() {
		final Configuration configuration = new Configuration();
		configuration.setProperty(CONFIG_PROPERTY_NAME_TIMEUNIT, this.timeunit.name());
		configuration.setProperty(CONFIG_PROPERTY_NAME_MAX_COLLECTION_DURATION, String.valueOf(this.maxCollectionDuration));
		configuration.setProperty(CONFIG_PROPERTY_NAME_MAX_DEVIATION, String.valueOf(this.maxDeviation));
		return configuration;
	}

	/**
	 * Buffer for similar traces that are to be agglomerated into a single trace.
	 * 
	 * @author Florian Biß
	 */
	private static final class TraceAgglomerationBuffer {
		private static final Log LOG = LogFactory.getLog(TraceAgglomerationBuffer.class);

		/** Contains all buffered traces for statistical purposes. */
		private final List<WorkflowTrace> traces = new ArrayList<WorkflowTrace>();

		/** Contains each unique trace at most once. */
		private WorkflowTrace agglomeratedTraces;

		private final long bufferCreatedTimestamp;

		/**
		 * Maximal runtime deviation factor. Use negative number to agglomerate all traces.
		 * 
		 * Outliers are indicated by <code>averageRuntime - deviationFactor * standardDeviation > runtime</code> or
		 * <code>runtime > averageRuntime + deviationFactor * standardDeviation</code>
		 * 
		 */
		private final long deviationFactor;

		/**
		 * Creates a new instance of this class.
		 */
		public TraceAgglomerationBuffer(final long bufferCreatedTimestamp, final long maxDeviation) {
			this.bufferCreatedTimestamp = bufferCreatedTimestamp;
			this.deviationFactor = maxDeviation;
		}

		/**
		 * Insert a trace into this buffer.
		 * 
		 * @param Trace
		 *            to insert
		 */
		public void insertTrace(final WorkflowTrace trace) {
			if (LOG.isDebugEnabled()) {
				LOG.debug("Inserting into AgglomerationBuffer: " + trace.toString());
			}
			synchronized (this) {
				this.traces.add(trace);
			}
		}

		/**
		 * Agglomerate all traces that can and should be agglomerated.
		 * 
		 * @return List of agglomerated traces and statistic outliers.
		 */
		public List<WorkflowTrace> processBuffer() {
			final List<WorkflowTrace> processed = new ArrayList<WorkflowTrace>();
			final StatisticInformation tmpRuntime;

			synchronized (this) {
				tmpRuntime = this.getTempBufferStatistic();
				for (final WorkflowTrace trace : this.traces) {

					// If deviationFactor is negative do not care about outliers.
					if ((this.deviationFactor > 0) && this.isOutlier(tmpRuntime, trace.getRuntime())) {
						if (LOG.isDebugEnabled()) {
							LOG.debug("Statistical outlier: " + trace.toString());
						}
						// Outlier. Do not agglomerate.
						processed.add(trace);
					} else {
						// Add to agglomeratedTraces.
						this.agglomerate(trace);
					}

				}

				processed.add(this.agglomeratedTraces);
			}

			return processed;
		}

		/**
		 * Merge a trace with the agglomerated traces.
		 * 
		 * @param trace
		 *            Trace to agglomerate
		 */
		private void agglomerate(final WorkflowTrace trace) {
			if (this.agglomeratedTraces == null) {
				// Trace is new, add to map.
				this.agglomeratedTraces = trace;
			} else {
				// Trace exists, merge statistics for every record.
				final WorkflowRecord[] agglomeratedRecords = this.agglomeratedTraces.getTraceEvents();
				final WorkflowRecord[] records = trace.getTraceEvents();
				for (int i = 0; i < agglomeratedRecords.length; i++) {
					agglomeratedRecords[i].getRuntime().merge(records[i].getRuntime());
				}

				this.agglomeratedTraces.getRuntime().merge(trace.getRuntime());
			}
		}

		/**
		 * Detect outliers.
		 * 
		 * @param averagedRuntimes
		 *            Statistics to check against
		 * @param recordRuntime
		 *            Data to check
		 * @return
		 *         <code>true</code> if both are too different
		 */
		private boolean isOutlier(final StatisticInformation averagedRuntimes, final StatisticInformation recordRuntime) {
			final long maxDeviation = averagedRuntimes.getStandardDeviation() * this.deviationFactor;
			if (maxDeviation < 0) {
				return true;
			}
			return Math.abs(recordRuntime.getAvg() - averagedRuntimes.getAvg()) > maxDeviation;
		}

		/**
		 * @return Creation time of this buffer.
		 */
		public long getBufferCreatedTimestamp() {
			synchronized (this) {
				return this.bufferCreatedTimestamp;
			}
		}

		/**
		 * The the statistical runtime informations about a buffer so far.
		 * 
		 * @return Runtime informations.
		 */
		private StatisticInformation getTempBufferStatistic() {
			final StatisticInformation tmpStatistic = new StatisticInformation();

			for (final WorkflowTrace trace : this.traces) {
				tmpStatistic.merge(trace.getRuntime());

			}
			return tmpStatistic;
		}
	}

	/**
	 * Compares traces based on their content instead of traceIds.
	 * Order of comparisons: Process name -> Trace lenght -> Id of each node in trace.
	 * If it all matches both traces are similar and may be agglomerated, albeit TraceId and runtimes might differ.
	 * 
	 * @author Florian Biß
	 * 
	 */
	private static final class TraceComperator implements Comparator<WorkflowTrace>, Serializable {
		private static final long serialVersionUID = 8920766818232517L;
		private static final Log LOG = LogFactory.getLog(TraceComperator.class);

		/**
		 * Creates a new instance of this class.
		 */
		public TraceComperator() {
			// default empty constructor
		}

		/**
		 * {@inheritDoc}
		 */
		public int compare(final WorkflowTrace t1, final WorkflowTrace t2) {
			final int compProcesses = t1.getProcessName().compareTo(t2.getProcessName());
			if (compProcesses != 0) {
				return compProcesses;
			}

			final WorkflowRecord[] recordsT1 = t1.getTraceEvents();
			final WorkflowRecord[] recordsT2 = t2.getTraceEvents();

			if ((recordsT1.length - recordsT2.length) != 0) {
				return recordsT1.length - recordsT2.length;
			}

			// Records in traces are already sorted by orderIndex, only compare nodeIds.
			for (int i = 0; i < recordsT1.length; i++) {
				final WorkflowRecord r1 = recordsT1[i];
				final WorkflowRecord r2 = recordsT2[i];
				final long idDiff = r1.getNodeId() - r2.getNodeId();

				if ((idDiff > Integer.MAX_VALUE) || (idDiff < Integer.MIN_VALUE)) {
					LOG.warn("Overflow during thread comparison!");
				}

				if (idDiff != 0) {
					return (int) (r1.getNodeId() - r2.getNodeId());
				}
			}

			// All records match.
			return 0;
		}
	}

}
